@@ -1,6 +1,6 @@
 /* app-nks.c - The Telesec NKS card application.
  * Copyright (C) 2004, 2007-2009 Free Software Foundation, Inc.
- * Copyright (C) 2004, 2007-2009, 2013-2015, 2020 g10 Code GmbH
+ * Copyright (C) 2004, 2007-2009, 2013-2015, 2020, 2022 g10 Code GmbH
  *
  * This file is part of GnuPG.
  *
@@ -100,7 +100,8 @@ static char const aid_idlm[] = { 0xD2, 0x76, 0x00, 0x00, 0x03, 0x0c, 0x01 };
 
 static struct
 {
-  int nks_app_id;/* One of the NKS_APP_ constants.  */
+  int nks_app_id;/* One of NKS_APP_*.  Keep them sorted so that no
+                  * unnecessary application switching is needed.  */
   int fid;       /* File ID. */
   int nks_ver;   /* 0 for NKS version 2, 3 for version 3, etc.  */
   int certtype;  /* Type of certificate or 0 if it is not a certificate. */
@@ -157,7 +158,7 @@ static struct
   { 1, 0xC008, 3, 101 },                    /* EF_C.CA.SIG  */
   { 1, 0xC00E, 3, 111 },                    /* EF_C.RCA.SIG  */
 
-  { 2, 0x4531, 15, 0, 0xC001, 0,1,1, 0x84}, /* EF_PK.CH.SIG  */
+  { 2, 0x4531, 15, 0, 0xC001, 0,1,0, 0x84}, /* EF_PK.CH.SIG  */
   { 2, 0xC000, 15,101 },                    /* EF.C.SCA.QES (SubCA) */
   { 2, 0xC001, 15,100 },                    /* EF.C.ICC.QES (Cert)  */
   { 2, 0xC00E, 15,111 },                    /* EF.C.RCA.QES (RootCA */
@@ -211,6 +212,8 @@ static gpg_error_t verify_pin (app_t app, int pwid, const char *desc,
                                gpg_error_t (*pincb)(void*, const char *,
                                                     char **),
                                void *pincb_arg);
+static gpg_error_t parse_keyref (app_t app, const char *keyref,
+                                 int want_keypair, int *r_fididx);
 
 
 
@@ -447,6 +450,8 @@ keygripstr_from_pk_file (app_t app, int pkfid, int cfid, char *r_gripstr,
 
   if (!err)
     {
+      if (r_algo)
+        *r_algo = algo;
       if (r_algostr)
         {
           *r_algostr = algostr;
@@ -500,7 +505,7 @@ static gpg_error_t
 find_fid_by_keyref (app_t app, const char *keyref, int *r_idx, int *r_algo)
 {
   gpg_error_t err;
-  int idx, fid, nks_app_id;
+  int idx;
   char keygripstr[2*KEYGRIP_LEN+1];
 
   if (!keyref || !keyref[0])
@@ -509,10 +514,14 @@ find_fid_by_keyref (app_t app, const char *keyref, int *r_idx, int *r_algo)
     {
       struct fid_cache_s *ci;
 
-      for (ci = app->app_local->fid_cache; ci; ci = ci->next)
+      /* FIXME: Our cache structure needs to be revised.  It doesn't
+       * take the app_id into account and we don't have a way to
+       * directly access the FID item if there are several of them
+       * with different app_ids.  We disable the cache for now. */
+      for (ci = app->app_local->fid_cache ; ci; ci = ci->next)
         if (ci->fid && ci->got_keygrip && !strcmp (ci->keygripstr, keyref))
           break;
-      if (ci) /* Cached */
+      if (ci && 0 ) /* Cached (disabled) */
         {
           for (idx=0; filelist[idx].fid; idx++)
             if (filelist[idx].fid == ci->fid)
@@ -577,45 +586,12 @@ find_fid_by_keyref (app_t app, const char *keyref, int *r_idx, int *r_algo)
     }
   else /* This is a usual keyref.  */
     {
-      if (!ascii_strncasecmp (keyref, "NKS-NKS3.", 9))
-        nks_app_id = NKS_APP_NKS;
-      else if (!ascii_strncasecmp (keyref, "NKS-ESIGN.", 10)
-               && app->app_local->qes_app_id == NKS_APP_ESIGN)
-        nks_app_id = NKS_APP_ESIGN;
-      else if (!ascii_strncasecmp (keyref, "NKS-SIGG.", 9)
-               && app->app_local->qes_app_id == NKS_APP_SIGG)
-        nks_app_id = NKS_APP_SIGG;
-      else if (!ascii_strncasecmp (keyref, "NKS-IDLM.", 9))
-        nks_app_id = NKS_APP_IDLM;
-      else if (!ascii_strncasecmp (keyref, "NKS-DF01.", 9))
-        nks_app_id = NKS_APP_NKS;
-      else
-        {
-          err = gpg_error (GPG_ERR_INV_ID);
-          goto leave;
-        }
-      keyref += nks_app_id == NKS_APP_ESIGN? 10 : 9;
-
-      if (!hexdigitp (keyref) || !hexdigitp (keyref+1)
-          || !hexdigitp (keyref+2) || !hexdigitp (keyref+3)
-          || keyref[4])
-        {
-          err = gpg_error (GPG_ERR_INV_ID);
-          goto leave;
-        }
-      fid = xtoi_4 (keyref);
-      for (idx=0; filelist[idx].fid; idx++)
-        if (filelist[idx].iskeypair && filelist[idx].fid == fid
-            && filelist[idx].nks_app_id == nks_app_id)
-          break;
-      if (!filelist[idx].fid)
-        {
-          err = gpg_error (GPG_ERR_NOT_FOUND);
-          goto leave;
-        }
+      err = parse_keyref (app, keyref, 1, &idx);
+      if (err)
+        goto leave;
       *r_idx = idx;
 
-      err = switch_application (app, nks_app_id);
+      err = switch_application (app, filelist[idx].nks_app_id);
       if (err)
         goto leave;
       if (r_algo)
@@ -795,6 +771,51 @@ do_getattr (app_t app, ctrl_t ctrl, const char *name)
 }
 
 
+/* Parse a keyref (NKS_*.*) and return the corresponding EF as an
+ * index into the filetable.  With WANT_KEYPAIR set a keypair EF is
+ * requested; otherwise also cert EFs are returned.  */
+static gpg_error_t
+parse_keyref (app_t app, const char *keyref, int want_keypair, int *r_fididx)
+{
+  int nks_app_id, fid, idx;
+
+  if (!ascii_strncasecmp (keyref, "NKS-NKS3.", 9))
+    nks_app_id = NKS_APP_NKS;
+  else if (!ascii_strncasecmp (keyref, "NKS-ESIGN.", 10)
+           && (!want_keypair || app->app_local->qes_app_id == NKS_APP_ESIGN))
+    nks_app_id = NKS_APP_ESIGN;
+  else if (!ascii_strncasecmp (keyref, "NKS-SIGG.", 9)
+           && (!want_keypair || app->app_local->qes_app_id == NKS_APP_SIGG))
+    nks_app_id = NKS_APP_SIGG;
+  else if (!ascii_strncasecmp (keyref, "NKS-IDLM.", 9))
+    nks_app_id = NKS_APP_IDLM;
+  else if (!ascii_strncasecmp (keyref, "NKS-DF01.", 9))
+    nks_app_id = NKS_APP_NKS;
+  else
+    return gpg_error (GPG_ERR_INV_ID);
+
+  keyref += nks_app_id == NKS_APP_ESIGN? 10 : 9;
+
+  if (!hexdigitp (keyref) || !hexdigitp (keyref+1)
+      || !hexdigitp (keyref+2) || !hexdigitp (keyref+3)
+      || keyref[4])
+    return gpg_error (GPG_ERR_INV_ID);
+  fid = xtoi_4 (keyref);
+  for (idx=0; filelist[idx].fid; idx++)
+    if (filelist[idx].fid == fid
+        && filelist[idx].nks_app_id == nks_app_id
+        && ((want_keypair && filelist[idx].iskeypair)
+            || (!want_keypair
+                && (filelist[idx].certtype || filelist[idx].iskeypair > 0))))
+      break;
+  if (!filelist[idx].fid)
+    return gpg_error (GPG_ERR_NOT_FOUND);
+
+  *r_fididx = idx;
+  return 0;
+}
+
+
 const char *
 get_nks_tag (app_t app, int nks_app_id)
 {
@@ -814,6 +835,21 @@ get_nks_tag (app_t app, int nks_app_id)
   return tag;
 }
 
+static void
+set_usage_string (char usagebuf[5], int i)
+{
+  int usageidx = 0;
+  if (filelist[i].issignkey)
+    usagebuf[usageidx++] = 's';
+  if (filelist[i].isauthkey)
+    usagebuf[usageidx++] = 'a';
+  if (filelist[i].isencrkey)
+    usagebuf[usageidx++] = 'e';
+  if (!usageidx)
+    usagebuf[usageidx++] = '-';
+  usagebuf[usageidx] = 0;
+}
+
 static void
 do_learn_status_core (app_t app, ctrl_t ctrl, unsigned int flags,
                       int nks_app_id)
@@ -856,7 +892,6 @@ do_learn_status_core (app_t app, ctrl_t ctrl, unsigned int flags,
         {
           char gripstr[40+1];
           char usagebuf[5];
-          int usageidx = 0;
           char *algostr = NULL;
 
           err = keygripstr_from_pk_file (app, filelist[i].fid,
@@ -869,15 +904,7 @@ do_learn_status_core (app_t app, ctrl_t ctrl, unsigned int flags,
             {
               snprintf (id_buf, sizeof id_buf, "NKS-%s.%04X",
                         tag, filelist[i].fid);
-              if (filelist[i].issignkey)
-                usagebuf[usageidx++] = 's';
-              if (filelist[i].isauthkey)
-                usagebuf[usageidx++] = 'a';
-              if (filelist[i].isencrkey)
-                usagebuf[usageidx++] = 'e';
-              if (!usageidx)
-                usagebuf[usageidx++] = '-';
-              usagebuf[usageidx] = 0;
+              set_usage_string (usagebuf, i);
               send_status_info (ctrl, "KEYPAIRINFO",
                                 gripstr, 40,
                                 id_buf, strlen (id_buf),
@@ -918,9 +945,8 @@ do_learn_status (app_t app, ctrl_t ctrl, unsigned int flags)
 }
 
 
-
 /* Helper to read a certificate from the file FID.  The function
- * assumes that the the application has already been selected.  */
+ * assumes that the application has already been selected.  */
 static gpg_error_t
 readcert_from_ef (app_t app, int fid, unsigned char **cert, size_t *certlen)
 {
@@ -954,7 +980,7 @@ readcert_from_ef (app_t app, int fid, unsigned char **cert, size_t *certlen)
       return err;
     }
 
-  if (!buflen || *buffer == 0xff)
+  if (!buflen || *buffer == 0xff || all_zero_p (buffer, buflen))
     {
       log_info ("nks: no certificate contained in FID 0x%04X\n", fid);
       err = gpg_error (GPG_ERR_NOT_FOUND);
@@ -1045,21 +1071,29 @@ iterate_over_filelist (app_t app, const char *want_keygripstr, int capability,
   for (idx++; filelist[idx].fid; idx++)
     {
       if (filelist[idx].nks_ver > app->appversion)
-        continue;
+        continue; /* EF not support by this card version.  */
 
       if (!filelist[idx].iskeypair)
-        continue;
+        continue; /* Skip - We are only interested in keypairs.  */
 
       if (app->app_local->only_idlm)
         {
+          /* IDLM cards have no other applications we want to switch
+           * to.  We skip all EFs which are not known for IDLM.  */
           if (filelist[idx].nks_app_id != NKS_APP_IDLM)
             continue;
         }
       else
         {
+          /* Skip all EFs which are not for NKS or the card's
+           * implementation for a qualified electoric signature (QES)
+           * which is either the old SIGG or the newer ESIGN.  */
           if (filelist[idx].nks_app_id != NKS_APP_NKS
               && filelist[idx].nks_app_id != app->app_local->qes_app_id)
             continue;
+
+          /* Switch if needed.  Note that the filelist should be
+           * sorted to avoid unnecessary switches.  */
           err = switch_application (app, filelist[idx].nks_app_id);
           if (err)
             {
@@ -1068,6 +1102,8 @@ iterate_over_filelist (app_t app, const char *want_keygripstr, int capability,
             }
         }
 
+      /* Get the keygrip from the EF.  Note that this functions
+       * consults the cache to avoid computing the keygrip again.  */
       err = keygripstr_from_pk_file (app, filelist[idx].fid,
                                      filelist[idx].iskeypair, keygripstr,
                                      NULL, NULL);
@@ -1080,6 +1116,7 @@ iterate_over_filelist (app_t app, const char *want_keygripstr, int capability,
 
       if (want_keygripstr)
         {
+          /* If the keygrip matches the requested one we are ready.  */
           if (!strcmp (keygripstr, want_keygripstr))
             {
               /* Found */
@@ -1087,8 +1124,10 @@ iterate_over_filelist (app_t app, const char *want_keygripstr, int capability,
               return 0;
             }
         }
-      else
+      else /* No keygrip requested - list all .  */
         {
+          /* If a capability has been requested return only keys with
+           * that capability.  */
           if (capability == GCRY_PK_USAGE_SIGN)
             {
               if (!filelist[idx].issignkey)
@@ -1105,7 +1144,7 @@ iterate_over_filelist (app_t app, const char *want_keygripstr, int capability,
                 continue;
             }
 
-          /* Found */
+          /* Found.  Return but save the last idenx of the loop.  */
           *idx_p = idx;
           return 0;
         }
@@ -1119,6 +1158,7 @@ iterate_over_filelist (app_t app, const char *want_keygripstr, int capability,
   return err;
 }
 
+
 /* Read the certificate with id CERTID (as returned by learn_status in
    the CERTINFO status lines) and return it in the freshly allocated
    buffer put into CERT and the length of the certificate put into
@@ -1127,69 +1167,46 @@ static gpg_error_t
 do_readcert (app_t app, const char *certid,
              unsigned char **cert, size_t *certlen)
 {
-  int i, fid;
+  int idx, fid;
   gpg_error_t err;
-  int nks_app_id;
 
   *cert = NULL;
   *certlen = 0;
 
-  /* Handle the case with KEYGRIP.  */
-  if (strlen (certid) == 40)
+  /* Handle the case with KEYGRIP.  We got a keygrip if the string has
+   * a length of 40 and does not start with an N as in NKS-* */
+  if (certid[0] != 'N' && strlen (certid) == 40)
     {
       char keygripstr[2*KEYGRIP_LEN+1];
 
-      i = -1;
-      err = iterate_over_filelist (app, certid, 0, keygripstr, &i);
+      idx = -1;
+      err = iterate_over_filelist (app, certid, 0, keygripstr, &idx);
       if (err)
         return err;
 
-      if (filelist[i].iskeypair > 0)
-        fid = filelist[i].iskeypair;
-      else
-        fid = filelist[i].fid;
-
-      return readcert_from_ef (app, fid, cert, certlen);
+      /* Switching is not required here because iterate_over_filelist
+       * has already done that.  */
     }
+  else /* This is not a keygrip.  */
+    {
+      err = parse_keyref (app, certid, 0, &idx);
+      if (err)
+        return err;
 
-  if (!strncmp (certid, "NKS-NKS3.", 9))
-    nks_app_id = NKS_APP_NKS;
-  else if (!strncmp (certid, "NKS-ESIGN.", 10))
-    nks_app_id = NKS_APP_ESIGN;
-  else if (!strncmp (certid, "NKS-SIGG.", 9))
-    nks_app_id = NKS_APP_SIGG;
-  else if (!strncmp (certid, "NKS-DF01.", 9))
-    nks_app_id = NKS_APP_NKS;
-  else if (!strncmp (certid, "NKS-IDLM.", 9))
-    nks_app_id = NKS_APP_IDLM;
-  else
-    return gpg_error (GPG_ERR_INV_ID);
-  certid += nks_app_id == NKS_APP_ESIGN? 10 : 9;
-
-  err = switch_application (app, nks_app_id);
-  if (err)
-    return err;
-
-  if (!hexdigitp (certid) || !hexdigitp (certid+1)
-      || !hexdigitp (certid+2) || !hexdigitp (certid+3)
-      || certid[4])
-    return gpg_error (GPG_ERR_INV_ID);
-  fid = xtoi_4 (certid);
-  for (i=0; filelist[i].fid; i++)
-    if ((filelist[i].certtype || filelist[i].iskeypair > 0)
-        && filelist[i].nks_app_id == nks_app_id
-        && filelist[i].fid == fid)
-      break;
-  if (!filelist[i].fid)
-    return gpg_error (GPG_ERR_NOT_FOUND);
+      err = switch_application (app, filelist[idx].nks_app_id);
+      if (err)
+        return err;
+    }
 
   /* If the requested objects is a plain public key, redirect it to
      the corresponding certificate.  The whole system is a bit messy
      because we sometime use the key directly or let the caller
      retrieve the key from the certificate.  The rationale for
      that is to support not-yet stored certificates. */
-  if (filelist[i].iskeypair > 0)
-    fid = filelist[i].iskeypair;
+  if (filelist[idx].iskeypair > 0)
+    fid = filelist[idx].iskeypair;
+  else
+    fid = filelist[idx].fid;
 
   return readcert_from_ef (app, fid, cert, certlen);
 }
@@ -1265,7 +1282,7 @@ do_readkey (app_t app, ctrl_t ctrl, const char *keyid, unsigned int flags,
       xfree (buffer[0]);
       xfree (buffer[1]);
     }
-  else if (strlen (keyid) == 40)
+  else if (keyid[0] != 'N' && strlen (keyid) == 40)
     {
       char keygripstr[2*KEYGRIP_LEN+1];
       int i = -1;
@@ -1280,7 +1297,6 @@ do_readkey (app_t app, ctrl_t ctrl, const char *keyid, unsigned int flags,
         {
           char *algostr;
           char usagebuf[5];
-          int usageidx = 0;
           char id_buf[100];
 
           if (app_help_get_keygrip_string_pk (*pk, *pklen, NULL, NULL, NULL,
@@ -1290,15 +1306,7 @@ do_readkey (app_t app, ctrl_t ctrl, const char *keyid, unsigned int flags,
           snprintf (id_buf, sizeof id_buf, "NKS-%s.%04X",
                     get_nks_tag (app, filelist[i].nks_app_id),
                     filelist[i].fid);
-          if (filelist[i].issignkey)
-            usagebuf[usageidx++] = 's';
-          if (filelist[i].isauthkey)
-            usagebuf[usageidx++] = 'a';
-          if (filelist[i].isencrkey)
-            usagebuf[usageidx++] = 'e';
-          if (!usageidx)
-            usagebuf[usageidx++] = '-';
-          usagebuf[usageidx] = 0;
+          set_usage_string (usagebuf, i);
           send_status_info (ctrl, "KEYPAIRINFO",
                             keygripstr, strlen (keygripstr),
                             id_buf, strlen (id_buf),
@@ -1390,10 +1398,9 @@ do_writecert (app_t app, ctrl_t ctrl,
    * writecert it won't harm to flush the entire cache.  */
   flush_fid_cache (app);
 
-
   /* The certificates we support all require PW1.CH.  Note that we
    * check that the nks_app_id matches which sorts out CERTID values
-   * which are subkecy to a different nks_app_id.  */
+   * which are subkeys to a different nks_app_id.  */
   desc = parse_pwidstr (app, "PW1.CH", 0, &tmp_app_id, &pwid);
   if (!desc || tmp_app_id != nks_app_id)
     return gpg_error (GPG_ERR_INV_ID);
@@ -1401,7 +1408,7 @@ do_writecert (app_t app, ctrl_t ctrl,
   if (err)
     return err;
 
-    /* Select the file and write the certificate.  */
+  /* Select the file and write the certificate.  */
   err = iso7816_select_file (app_get_slot (app), fid, 0);
   if (err)
     {
@@ -1606,7 +1613,7 @@ verify_pin (app_t app, int pwid, const char *desc,
   memset (&pininfo, 0, sizeof pininfo);
   pininfo.fixedlen = -1;
 
-  /* FIXME: TCOS allows to read the min. and max. values - do this.  */
+  /* FIXME: TCOS allows one to read the min. and max. values - do this.  */
   if (app->appversion == 15)
     {
       if (app->app_local->active_nks_app == NKS_APP_NKS && pwid == 0x03)
@@ -1743,28 +1750,32 @@ do_sign (app_t app, ctrl_t ctrl, const char *keyidstr, int hashalgo,
   unsigned char data[83];   /* Must be large enough for a SHA-1 digest
                                + the largest OID prefix. */
   size_t datalen;
+  int algo;
+  unsigned int digestlen;   /* Length of the hash.  */
+  unsigned char oidbuf[64];
+  size_t oidbuflen;
 
   (void)ctrl;
 
   switch (indatalen)
     {
-    case 20: // plain SHA-1 or RMD160 digest
-    case 28: // plain SHA-224 digest
-    case 32: // plain SHA-256 digest
-    case 48: // plain SHA-384 digest
-    case 64: // plain SHA-512 digest
-    case 35: // ASN.1 encoded SHA-1 or RMD160 digest
-    case 47: // ASN.1 encoded SHA-224 digest
-    case 51: // ASN.1 encoded SHA-256 digest
-    case 67: // ASN.1 encoded SHA-384 digest
-    case 83: // ASN.1 encoded SHA-512 digest
-        break;
+    case 20: /* plain SHA-1 or RMD160 digest         */
+    case 28: /* plain SHA-224 digest                 */
+    case 32: /* plain SHA-256 digest                 */
+    case 48: /* plain SHA-384 digest                 */
+    case 64: /* plain SHA-512 digest                 */
+    case 35: /* ASN.1 encoded SHA-1 or RMD160 digest */
+    case 47: /* ASN.1 encoded SHA-224 digest         */
+    case 51: /* ASN.1 encoded SHA-256 digest         */
+    case 67: /* ASN.1 encoded SHA-384 digest         */
+    case 83: /* ASN.1 encoded SHA-512 digest         */
+      break;
     default:
-      log_debug ("invalid length of input data: %zu\n", indatalen);
+      log_info ("nks: invalid length of input data: %zu\n", indatalen);
       return gpg_error (GPG_ERR_INV_VALUE);
     }
 
-  err = find_fid_by_keyref (app, keyidstr, &idx, NULL);
+  err = find_fid_by_keyref (app, keyidstr, &idx, &algo);
   if (err)
     return err;
 
@@ -1783,12 +1794,57 @@ do_sign (app_t app, ctrl_t ctrl, const char *keyidstr, int hashalgo,
 
   kid = filelist[idx].kid;
 
-  /* Prepare the DER object from INDATA.  */
-  if (app->appversion > 2 && (indatalen == 35
-                              || indatalen == 47
-                              || indatalen == 51
-                              || indatalen == 67
-                              || indatalen == 83))
+  digestlen = gcry_md_get_algo_dlen (hashalgo);
+
+  /* Prepare the input object from INDATA.  */
+  if (algo == GCRY_PK_ECC)
+    {
+      if (digestlen != 32 && digestlen != 48 && digestlen != 64)
+        {
+          log_error ("nks: ECC signing not possible: dlen=%u\n", digestlen);
+          return gpg_error (GPG_ERR_DIGEST_ALGO);
+        }
+
+      if (indatalen == digestlen)
+        {
+          /* Already prepared.  */
+          datalen = indatalen;
+          log_assert (datalen <= sizeof data);
+          memcpy (data, indata, datalen);
+        }
+      else if (indatalen > digestlen)
+        {
+          /* Assume a PKCS#1 prefix and remove it.  */
+          oidbuflen = sizeof oidbuf;
+          err = gcry_md_get_asnoid (hashalgo, &oidbuf, &oidbuflen);
+          if (err)
+            {
+              log_error ("nks: no OID for hash algo %d\n", hashalgo);
+              return gpg_error (GPG_ERR_INTERNAL);
+            }
+          if (indatalen != oidbuflen + digestlen
+              || memcmp (indata, oidbuf, oidbuflen))
+            {
+              log_error ("nks: input data too long for ECC: len=%zu\n",
+                         indatalen);
+              return gpg_error (GPG_ERR_INV_VALUE);
+            }
+          datalen = indatalen - oidbuflen;
+          log_assert (datalen <= sizeof data);
+          memcpy (data, (const char*)indata + oidbuflen, datalen);
+        }
+      else
+        {
+          log_error ("nks: input data too short for ECC: len=%zu\n",
+                     indatalen);
+          return gpg_error (GPG_ERR_INV_VALUE);
+        }
+    }
+  else if (app->appversion > 2 && (indatalen == 35
+                                   || indatalen == 47
+                                   || indatalen == 51
+                                   || indatalen == 67
+                                   || indatalen == 83))
     {
       /* Verify that the caller has sent a proper ASN.1 encoded hash
          for RMD160 or SHA-{1,224,256,384,512}. */
@@ -1824,7 +1880,8 @@ do_sign (app_t app, ctrl_t ctrl, const char *keyidstr, int hashalgo,
       memcpy (data, indata, indatalen);
       datalen = 35;
     }
-  /* Concatenate prefix and digest.  */
+  /* Concatenate prefix and digest.
+   * Note that the X macro creates an "else if".  Ugly - I know. */
 #define X(algo,prefix,plaindigestlen) \
   if ((hashalgo == (algo)) && (indatalen == (plaindigestlen))) \
     {                                                          \
@@ -1843,23 +1900,36 @@ do_sign (app_t app, ctrl_t ctrl, const char *keyidstr, int hashalgo,
     return gpg_error (GPG_ERR_INV_VALUE);
 #undef X
 
-  /* Send an MSE for PSO:Computer_Signature.  */
-  if (app->appversion > 2)
+  /* Send an MSE for PSO:Compute_Signature.  */
+  if (app->appversion > 2 && app->app_local->active_nks_app != NKS_APP_ESIGN)
     {
       unsigned char mse[6];
+      unsigned int mselen;
 
-      mse[0] = 0x80; /* Algorithm reference.  */
-      mse[1] = 1;
-      mse[2] = 2;    /* RSA, card does pkcs#1 v1.5 padding, no ASN.1 check.  */
-      mse[3] = 0x84; /* Private key reference.  */
-      mse[4] = 1;
-      mse[5] = kid;
+      if (algo == GCRY_PK_ECC)
+        {
+          mse[0] = 0x84; /* Private key reference.  */
+          mse[1] = 1;
+          mse[2] = kid;
+          mselen = 3;
+        }
+      else /* RSA */
+        {
+          mse[0] = 0x80; /* Algorithm reference.  */
+          mse[1] = 1;
+          mse[2] = 2;    /* Card does pkcs#1 v1.5 padding, no ASN.1 check.  */
+          mse[3] = 0x84; /* Private key reference.  */
+          mse[4] = 1;
+          mse[5] = kid;
+          mselen = 6;
+        }
       err = iso7816_manage_security_env (app_get_slot (app), 0x41, 0xB6,
-                                         mse, sizeof mse);
+                                         mse, mselen);
     }
 
-  /* We use the Global PIN 1 */
-  if (app->appversion == 15)
+  if (app->app_local->active_nks_app == NKS_APP_ESIGN)
+    pwid = 0x81;
+  else if (app->appversion == 15)
     pwid = 0x03;
   else
     pwid = 0x00;
@@ -2333,22 +2403,13 @@ do_with_keygrip (app_t app, ctrl_t ctrl, int action,
       else
         {
           char idbuf[20];
-          const char *tagstr;
-
-          if (app->app_local->active_nks_app == NKS_APP_ESIGN)
-            tagstr = "ESIGN";
-          else if (app->app_local->active_nks_app == NKS_APP_SIGG)
-            tagstr = "SIGG";
-          else if (app->app_local->active_nks_app == NKS_APP_IDLM)
-            tagstr = "IDLM";
-          else if (app->appversion < 3)
-            tagstr = "DF01";
-          else
-            tagstr = "NKS3";
+          char usagebuf[5];
 
           snprintf (idbuf, sizeof idbuf, "NKS-%s.%04X",
-                    tagstr, filelist[idx].fid);
-          send_keyinfo (ctrl, data, keygripstr, serialno, idbuf);
+                    get_nks_tag (app, app->app_local->active_nks_app),
+                    filelist[idx].fid);
+          set_usage_string (usagebuf, idx);
+          send_keyinfo (ctrl, data, keygripstr, serialno, idbuf, usagebuf);
         }
     }
 
